// $Id: DoctypeChangerStream.java,v 1.6 2002/04/15 01:02:03 jeff Exp $
/*
 * Copyright (c) 1999-2001 by Simon St.Laurent and Jeff Turner.  All Rights
 * Reserved.
 *
 * This program is open source software; you may use, copy, modify, and
 * redistribute it under the terms of the LICENSE with which it was
 * originally distributed.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * LICENSE for more details.
 */

package net.socialchange.doctype;

import java.io.InputStream;
import java.io.FilterInputStream;
import java.io.IOException;

/**
 * <p>This class adds, removes or otherwise modifies a DOCTYPE declaration to an
 * incoming XML document.</p>
 * <p>For examples of how to use this, see the <a
 * href="../../../overview-summary.html">package overview</a>.
 * </p>
 *
 * @author Simon St.Laurent
 * @author Nigel Whitaker
 * @author <a href="jeff@socialchange.net.au">Jeff Turner</a>
 * @version $Revision: 1.6 $  $Date: 2002/04/15 01:02:03 $
 */
public final class DoctypeChangerStream extends FilterInputStream
{
	// ----------------------------------------------------- Instance Variables

	protected Doctype doctype;
	protected DoctypeGenerator generator;

	protected boolean docStarted=false;
	protected boolean inComment=false;
	protected StringBuffer myBuffer=new StringBuffer();


	// --------------------------------------------------------- Public Methods

	/**
	 * Constructor.
	 * @param in The InputStream which will return XML/SGML.
	 */
	public DoctypeChangerStream (InputStream in) {
		super(in);
	}

	/**
	 * Set the Generator which will determine the policy for what to do with the
	 * DOCTYPE declaration. See the package overview for more details.
	 */
	public void setGenerator(final DoctypeGenerator generator) {
		this.generator = generator;
	}

	/**
	 * Get the Doctype set for the parsed stream. If no DOCTYPE declaration was
	 * set, this will be <code>null</code>. Likewise if you call this before
	 * doing the parse.
	 */
	public Doctype getDoctype() {
		return this.doctype;
	}

	public int read() throws IOException {
		int c =' ';
		if (myBuffer.length()==0) {
			c=in.read();
			if ((docStarted==false) && (inComment==false) && (c=='<')) {
				//figure out if we have a Doctype declaration
				int d=in.read();
				switch ( d ) {
					case ('?'):	//question mark, let it go
						myBuffer.append((char) d);
						break;

					case ('!'):	//either comment or Doctype
						int e=in.read();
						if (e=='D') {
							// Doctype! Bingo.

							DoctypeParser parser = new DoctypeParser();
							StringBuffer doctypeBuf = new StringBuffer(200);

							// add the chars we've read (not c, because it will
							// be returned by this function).
							doctypeBuf.append((char)d).append((char)e);
							parser.parse(in, doctypeBuf);
							if (generator == null) {
								// if the user forgot to set a generator, default to using the old doctype.
								this.doctype = parser.getDoctype();
							} else {
								this.doctype = generator.generate(parser.getDoctype());
							}

							if (this.doctype == null) {
								// Remove doctype
								doctypeBuf = null;

								// because c still contains '<', we need to
								// replace it with the next char.
								c = in.read();

							} else if (this.doctype.equals(parser.getDoctype())) {
								// use old doctype
								myBuffer.append(doctypeBuf.toString());

							} else if (this.doctype != null) {
								// Replace doctype
								doctypeBuf = null; // discard old doctype
								addDocType(); // add new doctype

							}
							docStarted=true;

						} else if (e=='-') {
							int f=in.read();
							if (f=='-') {
								// Comment.
								inComment=true;
								myBuffer.append((char) d);
								myBuffer.append((char) e);
								myBuffer.append((char) f);
							}

						} else {//e didn't equal 'D'
							myBuffer.append((char) d);
							myBuffer.append((char) e);
						} //end e=='D'
						break;

					default:	//root element, need to insert in front

						if (generator != null)
							this.doctype = generator.generate(null);

						if (this.doctype != null) {
							addDocType(); // add new doctype
							myBuffer.append('<');

							/* Why the above append? This is complicated.. 
							 * Notice how we *always* return the char in c?
							 * Well if we're in this 'if' block, c will be
							 * '<'. Furthermore, because we're in this 'default'
							 * switch area, that '<' was the first of the root
							 * element.
							 * So now we've added '!Doctype ... ' to the buffer.
							 * That means when we get out of this 'if', the c
							 * will be used for the first char of '<!Doctype..',
							 * and the '<' we append to the buffer here will be
							 * used for the root element.
							 */
						}
						myBuffer.append((char) d); // first char of the element name
						docStarted=true;
						break;

				}//end switch


			}//end c=='<', docStarted==false
			else if (inComment==true && c=='-') {
				int d=in.read();
				if (d=='-') {
					int e=in.read();
					if (e=='>') {
						inComment=false;
					}
					myBuffer.append((char) d);
					myBuffer.append((char) e);
				} else {
					myBuffer.append((char) d);
				}
			}
		} else {
			c=feedFromInternalBuffer();
		}
		return c;

	} 


	// --------------------------------------------------------- Private/Protected Methods

	/**
	  This method is a placeholder - all 'real' activity appears in the int
	  read() method.
	  This placeholder is substantially from Java I/O by Elliotte Rusty Harold,
http://www.oreilly.com/catalog/javaio/.
	 */
	private boolean endOfStream = false;

	public int read(byte[] text, int offset, int length) throws IOException {

		if (endOfStream) return -1;
		int numRead = 0;

		for (int i = offset; i < offset+length; i++) {
			int temp = this.read();
			if (temp == -1) {
				this.endOfStream = true;
				break;
			}
			text[i] = (byte) temp;
			numRead++;
		}
		return numRead;

	}

	/**
	 * Return the argument String, wrapped in quotes. Defaults to ' unless that
	 * char is present in <code>str</code>, in which case " is used.
	 */
	private void addQuotedStr(final String str) {
		char quote = (str.indexOf("'") == -1) ? '\'' : '"';
		myBuffer.append(quote);
		myBuffer.append(str);
		myBuffer.append(quote);
	}

	/**
	 * Add the DOCTYPE declaration to the buffer, from which it will be read in
	 * subsequent {@link #read} operations.
	 *
	 * We add some line separators here for readability.
	 */
	protected void addDocType(){
		String eol;
		try { eol = System.getProperty ("line.separator", "\n"); }
		catch (SecurityException e) { eol = "\n"; }

		final String root = this.doctype.getRootElement();
		final String sysId = this.doctype.getSystemId();
		final String pubId = this.doctype.getPublicId();
		final String intSubset = this.doctype.getInternalSubset();
		myBuffer.append("!DOCTYPE ");
		myBuffer.append(root);
		if (pubId == null) {
			if (sysId != null) {
				myBuffer.append(" SYSTEM ");
				addQuotedStr(sysId);
			}
		} else {
			myBuffer.append(" PUBLIC ");
			addQuotedStr(pubId);
			myBuffer.append(" ");
			addQuotedStr(sysId);
		}
		myBuffer.append(" ");

		if (intSubset != null) {
			myBuffer.append("["+eol);
			myBuffer.append(intSubset);
			myBuffer.append(eol+"]");
		}

		myBuffer.append(">"+eol);
		docStarted=true;

	}

	protected int feedFromInternalBuffer() {
		//feed out the buffer a character at a time
		int retChar=myBuffer.charAt(0);
		myBuffer.reverse();
		myBuffer.setLength(myBuffer.length()-1);
		myBuffer.reverse();
		return retChar;
	}

}
