// $Id: DoctypeParser.java,v 1.3 2001/11/14 11:47:42 jeff Exp $
/*
 * Copyright (c) 1999-2001 by Simon St.Laurent and Jeff Turner.  All Rights
 * Reserved.
 *
 * This program is open source software; you may use, copy, modify, and
 * redistribute it under the terms of the LICENSE with which it was
 * originally distributed.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * LICENSE for more details.
 */

package net.socialchange.doctype;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.IOException;

/**
 * Class for parsing a Doctype declaration and storing read information for
 * later retrieval. Straightforward state machine implementation.
 *
 * If the parse fails for any reason,
 * the {@link #parse(InputStream, StringBuffer)} method will silently return. You
 * can retrieve an error message through {@link #getErrorMessage}.
 *
 * Note: you can get this info through standard SAX and DOM methods. This is a
 * pre-parser so that we can implement conditional replacement of Doctype
 * declaration components.
 *
 * @author <a href="jeff@socialchange.net.au">Jeff Turner</a>
 * @version $Revision: 1.3 $ $Date: 2001/11/14 11:47:42 $
 */

public class DoctypeParser {

	// ----------------------------------------------------- Instance Variables
    /**
     * The debugging detail level for this component.
     */
    protected final int debug;

	private Doctype doctype;
	private String errMsg;

	// --------------------------------------------------------- Public Methods

	/**
	 * Instantiate a parser.
	 */
	public DoctypeParser() {
		this.debug = 0;
	}

	/**
	 * Instantiate a parser, with a certain debug level.
	 */
	public DoctypeParser(final int debugLevel) {
		this.debug = debugLevel;
	}

	/**
	 * Parse a doctype declaration.
	 *
	 * After this call, data may be retrieved with the accessor methods.
	 *
	 * If the parsed doctype is invalid, this method will silently return.
	 *
	 * @param in An InputStream where the next character will be somewhere in
	 * the text <code>&lt;!Doctype</code>.
	 * @see #getErrorMessage
	 */
	public void parse(final InputStream in) throws IOException {
		parse(in, null);
	}

	/**
	 * Parse a doctype declaration, storing read characters.
	 *
	 * After this call, data may be retrieved with the accessor methods.
	 *
	 * If the parsed doctype is invalid, this method will silently return.
	 *
	 * @param in An InputStream where the next character will be somewhere in
	 * the text <code>&lt;!Doctype</code>.
	 * @param buf The StringBuffer to store read characters to. If
	 * <code>null</code>, behave same as {@link #parse(InputStream)}.
	 *
	 * @see #getErrorMessage
	 */
	public void parse(final InputStream in, final StringBuffer buf) throws IOException {

		// Main state flags
		final char STATE_Doctype='d'; // we're parsing the text '<!Doctype'
		final char STATE_NAME='n'; // we're parsing the root element name
		final char STATE_EXTID='e'; // we're parsing the external id specifier, 'PUBLIC' or 'SYSTEM'
		final char STATE_SYS_LIT='s'; // we're parsing the system identifier
		final char STATE_PUB_LIT='p';  // we're parsing the public identifier
		final char STATE_INT_SUBSET='i';  // we're parsing the internal subset
		final char STATE_EXIT='x'; // we've reached the end of the doctype decl.
		final char STATE_FAIL='f'; // parsing error 

		// State modifier flags
		boolean whitespace = false; // in any state, clearing the initial whitespace
		boolean inQuote = false; // in sys or public literal state, inside a quote
		boolean isSingleQuote = false; // the abovementioned quote is a single '.
		boolean inSubset = false; // in internal subset state, we've entered the first '['

		int state = STATE_Doctype;

		// buffers to store read info
		final StringBuffer nameBuf = new StringBuffer(); // root element name
		final StringBuffer extidBuf = new StringBuffer(6); // 'PUBLIC' or 'SYSTEM'
		final StringBuffer syslitBuf = new StringBuffer(); // system identifier
		final StringBuffer publitBuf = new StringBuffer(); // public identifier
		final StringBuffer subsetBuf = new StringBuffer(); // internal subset

		char e;
		while (state != STATE_EXIT && state != STATE_FAIL) {
			e = (char)in.read();

			if ((short)e == -1) {
				this.errMsg = "Doctype decl parser reached end of stream";
				state = STATE_FAIL;
			}
			// if you don't typecast this, then a ByteArrayInputStream from a
			// String will return 65535, not -1. Don't know if this is a bug.. I'm
			// using standard Sun JDK 1.3.

			if (buf != null) buf.append(e);

			// loop on whitespace
			if (whitespace) {
				if (isWhitespace(e)) continue; else whitespace = false;
			}
			switch (state) {
				case (STATE_Doctype):
					if (isWhitespace(e)) {
						if (debug>1) log("Switched from STATE_Doctype to STATE_NAME");
						whitespace = true;
						state = STATE_NAME;
					}
					break;
				case (STATE_NAME):
					if (isWhitespace(e)) {
						if (debug>1) log("Switched from STATE_NAME to STATE_EXTID");
						whitespace = true;
						state = STATE_EXTID;
					} else {
						nameBuf.append(e);
					}
					break;

				case (STATE_EXTID):
					if (isWhitespace(e)) {
						whitespace = true;
						String extid = extidBuf.toString();
						if (debug>1) log("We have accumulated extid "+extid);
						if (extid.equals("PUBLIC")) {
							if (debug>1) log("Switched from STATE_EXTID to STATE_PUB_LIT");
							state = STATE_PUB_LIT;
						} else if (extid.equals("SYSTEM")) {
							if (debug>1) log("Switched from STATE_EXTID to STATE_SYS_LIT");
							state = STATE_SYS_LIT;
						} else {
							this.errMsg = "Doctype decl parser read unknown external ID string "+extid+". Should be SYSTEM or PUBLIC";
							state = STATE_FAIL;
						}
					} else {
						extidBuf.append(e);
					}
					break;

				case (STATE_PUB_LIT):
					if (e == '"' && inQuote == false) {
						if (debug>1) log("Entering \"");
						inQuote = true;
						isSingleQuote = false;
					} else {
						if (e == '\'' && inQuote == false) {
							if (debug>1) log("Entering '");
							inQuote = true;
							isSingleQuote = true;
						} else {
							if ( (e == '"' && inQuote == true && isSingleQuote == false) ||
									(e == '\'' && inQuote == true && isSingleQuote == true) ) {
								if (debug>1) log("Leaving "+(char)e);
								inQuote = false;
								whitespace = true;
								if (debug>1) log("Switched from STATE_PUB_LIT to STATE_SYS_LIT");
								state = STATE_SYS_LIT;
							} else {
								publitBuf.append(e);
							}
						}
					}
					break;

				case (STATE_SYS_LIT):
						if (e == '"' && inQuote == false) {
							if (debug>1) log("Entering \"");
							inQuote = true;
							isSingleQuote = false;
						} else {
							if (e == '\'' && inQuote == false) {
								if (debug>1) log("Entering '");
								inQuote = true;
								isSingleQuote = true;
							} else {
								if ( (e == '"' && inQuote == true && isSingleQuote == false) ||
										(e == '\'' && inQuote == true && isSingleQuote == true) ) {
									if (debug>1) log("Leaving "+(char)e);
									inQuote = false;
									whitespace = true;
									if (debug>1) log("Switched from STATE_SYS_LIT to STATE_INT_SUBSET");
									state = STATE_INT_SUBSET;
								} else {
									syslitBuf.append(e);
								}
							}
						}
						break;

				case (STATE_INT_SUBSET):
						if (inSubset == false) {
							if (e == '[') {
								inSubset = true;
							} else if (e == '>') {
								state = STATE_EXIT;
								if (debug>1) log("Switched from STATE_INT_SUBSET to STATE_EXIT");
							}
						} else {
							if (e == ']') {
								if (debug>1) log("Leaving ]");
								inSubset = false;
							} else {
								subsetBuf.append(e);
							}
						}
						break;
			}

		}
		final String pubId = publitBuf.toString();
		final String intSubset = subsetBuf.toString();
		final boolean noPubId = (pubId.length() == 0);
		final boolean noIntSubset = (intSubset.length() == 0);

		this.doctype = new DoctypeImpl(
				nameBuf.toString(),
				(noPubId?null:pubId),
				syslitBuf.toString(),
				(noIntSubset?null:intSubset)
				);
	}

	public final Doctype getDoctype() {
		return this.doctype;
	}

	/**
	 * If a parse failed, this will return any error message associated with the
	 * failure
	 */
	public String getErrorMessage() {
		return this.errMsg;
	}

	// --------------------------------------------------------- Private Methods
	/**
	 * Test if a character is whitespace.
	 * <pre>
	 * [3] S ::= (#x20 | #x9 | #xd | #xa)+
	 * </pre>
	 * Stolen from <code>com.icl.saxon.aelfred.XmlParser</code>.
	 * @param c The character to test.
	 * @return true if the character is whitespace.
	 */
	private final boolean isWhitespace (char c)
	{
		if (c > 0x20)
			return false;
		if (c == 0x20 || c == 0x0a || c == 0x09 || c == 0x0d)
			return true;
		return false;   // illegal ...
	}

	public void log(String s) {
		System.out.println(s);
	}

	public void log(int s) {
		System.out.print(""+(char)s);
	}

}
